﻿#include "monsters.h"
#include <QtXml/qdom.h>
#include <QFile>
#include <QDebug>
#include "../common/sources/Map.h"
#include "../common/sources/system/filecache.h"
#include "../common/sources/system/cgraphics2d.h"
#include "player.h"
#include "walls.h"
#include "interface.h"
#include "game.h"
#include "sound.h"

CMonsters    monsters;

//---------------------------------------------------------------------------------------------
CMonsters::CMonsters()
{

}

//---------------------------------------------------------------------------------------------
void CMonsters::readMonstersDB()
{
    QDomDocument doc;

    QFile f("databases/monsters.xml");
    f.open(QIODevice::ReadOnly);
    doc.setContent(&f);
    f.close();

    QDomElement root = doc.documentElement();
    QDomNode    monsterNode = root.firstChild();

    while(!monsterNode.isNull())
    {
        if (monsterNode.isElement())
        {
            QDomElement monster = monsterNode.toElement();
            std::vector<CParamType> paramsTypesList;

            if (monster.tagName() == "monster")
            {
                QDomElement monsterInfo = monster.firstChild().toElement();
                SMonsterData    monsterData;

                while(!monsterInfo.isNull())
                {
                    if (monsterInfo.tagName() == "img_front")
                    {
                        monsterData.frontImage = QString("gfx/3DView/monsters/") + monsterInfo.text();
                    }
                    else if (monsterInfo.tagName() == "img_side")
                    {
                        monsterData.sideImage = QString("gfx/3DView/monsters/") + monsterInfo.text();
                    }
                    else if (monsterInfo.tagName() == "img_back")
                    {
                        monsterData.backImage = QString("gfx/3DView/monsters/") + monsterInfo.text();
                    }
                    else if (monsterInfo.tagName() == "img_attack")
                    {
                        monsterData.attackImage = QString("gfx/3DView/monsters/") + monsterInfo.text();
                    }
                    else if (monsterInfo.tagName() == "time_between_atk")
                    {
                        monsterData.timeBetweenAttack = monsterInfo.text().toInt();
                    }
                    else if (monsterInfo.tagName() == "atk_display_time")
                    {
                        monsterData.attackDisplayTime = monsterInfo.text().toInt();
                    }
                    else if (monsterInfo.tagName() == "atk_sound")
                    {
                        monsterData.attackSound = QString("sound/") + monsterInfo.text();
                    }
                    else if (monsterInfo.tagName() == "param")
                    {
                        QString paramType = monsterInfo.attribute("type");

                        CParamType  newParamType;

                        if (paramType == "int")
                        {
                            newParamType.mType = eParamInt;
                            newParamType.mName = monsterInfo.text();
                            paramsTypesList.push_back(newParamType);
                        }
                        else if (paramType == "enum")
                        {
                            newParamType.mType = eParamEnum;
                            newParamType.mName = monsterInfo.text();
                            newParamType.mValues = monsterInfo.attribute("values").split(";");
                            paramsTypesList.push_back(newParamType);
                        }
                        else if (paramType == "stack")
                        {
                            newParamType.mType = eParamStack;
                            paramsTypesList.push_back(newParamType);
                        }
                    }

                    monsterInfo = monsterInfo.nextSibling().toElement();
                }

                CMap::mMonstersParams.push_back(paramsTypesList);
                monstersDatas.push_back(monsterData);
            }
        }
        monsterNode = monsterNode.nextSibling();
    }
}

//---------------------------------------------------------------------------------------------
void CMonsters::init()
{
    for (size_t i = 0; i < map.mMonsterGroups.size(); ++i)
    {
        CMonsterGroup&  mapGroup = map.mMonsterGroups[i];
        CMonsterGroup2	group;
        int nbMonsters = mapGroup.getIntParam("nbMonstres");

        group.dir = (EMonsterDir)mapGroup.getEnumParam("Dir");

        for (int j = 0; j < nbMonsters; j++)
            group.monsters[j].dir = group.dir;

        if (nbMonsters == 1)
        {
            group.monsters[0].pos = eMonsterPosCenter;
            group.monsters[0].life = 3;

            for (int j = 1; j < 4; j++)
            {
                group.monsters[j].pos = eMonsterPosNone;
                group.monsters[j].life = 0;
            }
        }
        else
        {
            // positions des monstres pour une direction donnée
            static const EMonsterPos startingPos[4][4] =
            {
                {eMonsterPosNorthWest, eMonsterPosNorthEast, eMonsterPosSouthWest, eMonsterPosSouthEast},   // up
                {eMonsterPosSouthEast, eMonsterPosSouthWest, eMonsterPosNorthEast, eMonsterPosNorthWest},   // down
                {eMonsterPosNorthWest, eMonsterPosSouthWest, eMonsterPosNorthEast, eMonsterPosSouthEast},   // left
                {eMonsterPosSouthEast, eMonsterPosNorthEast, eMonsterPosSouthWest, eMonsterPosNorthWest}    // right
            };

            for (int j = 0; j < 4; j++)
            {
                if (j < nbMonsters)
                {
                    group.monsters[j].pos = startingPos[group.dir][j];
                    group.monsters[j].life = 3;
                }
                else
                {
                    group.monsters[j].pos = eMonsterPosNone;
                    group.monsters[j].life = 0;
                }
            }
        }

        monsterGroups.push_back(group);
    }
}

//---------------------------------------------------------------------------------------------
CVec2 CMonsters::getMonsterPos(float x, float y)
{
    CVec2   pos;
    pos.x = 250 * (x - 4.5) / (8.5 - y) + 112;
    pos.y = 310 / (8.5 - 0.9 * y) + 66;
    return pos;
}

//---------------------------------------------------------------------------------------------
void CMonsters::draw(QImage* image, CVec2 mapPos, CVec2 tablePos)
{
    if (tablePos.y == WALL_TABLE_HEIGHT - 1)
        return;

    int groupIndex = map.findMonsterGroupIndex(mapPos);

    if (groupIndex != -1)
    {
        CMonsterGroup&  mapGroup = map.mMonsterGroups[groupIndex];
        CMonsterGroup2&	group = monsterGroups[groupIndex];

        // position des monstres en fonction de la position du joueur
        static const float monsterPosDir[4][5][2] =
        {
            //   NW          NE          SW          SE           C
            {{0.0, 0.0}, {1.0, 0.0}, {0.0, 1.0}, {1.0, 1.0},  {0.5, 0.5}},   // haut
            {{1.0, 0.0}, {1.0, 1.0}, {0.0, 0.0}, {0.0, 1.0},  {0.5, 0.5}},   // gauche
            {{1.0, 1.0}, {0.0, 1.0}, {1.0, 0.0}, {0.0, 0.0},  {0.5, 0.5}},   // bas
            {{0.0, 1.0}, {0.0, 0.0}, {1.0, 1.0}, {1.0, 0.0},  {0.5, 0.5}}    // droite
        };

        // direction des monstres en fonction de la direction du joueur
        static const EMonsterDir monsterDirDir[4][4] =
        {
            {eMonsterDirUp,    eMonsterDirDown,  eMonsterDirLeft,  eMonsterDirRight}, // haut
            {eMonsterDirRight, eMonsterDirLeft,  eMonsterDirUp,    eMonsterDirDown},  // gauche
            {eMonsterDirDown,  eMonsterDirUp,    eMonsterDirRight, eMonsterDirLeft},  // bas
            {eMonsterDirLeft,  eMonsterDirRight, eMonsterDirDown,  eMonsterDirUp}     // droite
        };

        // on dessine la ligne arrière des monstres avant ceux de devant
        for (int row = 0; row < 3; ++row)
        {
            for (int i = 0; i < 4; ++i)
            {
                CMonster&   monster = group.monsters[i];

                if (monster.pos != eMonsterPosNone)
                {
                    // direction relative
                    EMonsterDir tableDir = monsterDirDir[player.dir][monster.dir];

                    // sprite
                    int type = mapGroup.getType();
                    QString spriteName;
                    bool    flip = false;

                    if (tableDir == eMonsterDirDown)
                    {
                        // de face
                        if (monster.displayAttackTimer.isRunning())
                        {
                            // attaque
                            spriteName = monstersDatas[type].attackImage;
                            if (spriteName.isEmpty() == true)
                                spriteName = monstersDatas[type].frontImage;
                        }
                        else
                        {
                            // normal
                            spriteName = monstersDatas[type].frontImage;
                        }
                    }
                    else if (tableDir == eMonsterDirUp)
                    {
                        // de dos
                        spriteName = monstersDatas[type].backImage;
                        if (spriteName.isEmpty() == true)
                            spriteName = monstersDatas[type].frontImage;
                    }
                    else
                    {
                        // de coté
                        spriteName = monstersDatas[type].sideImage;
                        if (spriteName.isEmpty() == true)
                            spriteName = monstersDatas[type].frontImage;
                        else if (tableDir == eMonsterDirLeft)
                            flip = true;
                    }

                    if (monsterPosDir[player.dir][monster.pos - eMonsterPosNorthWest][1] == (float)row / 2.0)
                    {
                        // position dans la table
                        float tablePosX = tablePos.x * 2 + monsterPosDir[player.dir][monster.pos - eMonsterPosNorthWest][0];
                        float tablePosY = tablePos.y * 2 + monsterPosDir[player.dir][monster.pos - eMonsterPosNorthWest][1];

                        // position à l'écran
                        QImage  sprite = fileCache.getImage(spriteName.toStdString());
                        CVec2   pos = getMonsterPos(tablePosX, tablePosY);

                        // calcule la taille par rapport à la position de référence (la plus proche)
                        CVec2   pos0, pos1;
                        float   scale;

                        pos0 = getMonsterPos(WALL_TABLE_WIDTH, (WALL_TABLE_HEIGHT - 1) * 2 + 1);
                        pos1 = getMonsterPos(WALL_TABLE_WIDTH, tablePosY);
                        scale = 2.37 * (float)(pos1.x - 112) / (float)(pos0.x - 112);

                        // redimensionne le monstre dans une image temporaire
                        QImage  scaledMonster(sprite.size() * scale, QImage::Format_ARGB32);
                        scaledMonster.fill(TRANSPARENT);
                        graph2D.drawImageScaled(&scaledMonster, CVec2(), sprite, scale);

                        // assombrit le monstre en fonction de la distance
                        float shadow = ((WALL_TABLE_HEIGHT - 1) * 2 - tablePosY - 1) * 0.13f;
                        graph2D.darken(&scaledMonster, shadow);

                        // dessine le monstre (position relative à ses pieds)
                        pos.x -= scaledMonster.width() / 2;
                        pos.y -= scaledMonster.height() - 9 * scale;
                        graph2D.drawImage(image, pos, scaledMonster, 0, flip, QRect(0, 33, MAINVIEW_WIDTH, MAINVIEW_HEIGHT));
                    }
                }
            }
        }
    }
}

//---------------------------------------------------------------------------------------------
void CMonsters::update()
{
    for (size_t i = 0; i < map.mMonsterGroups.size(); ++i)
    {
        CMonsterGroup&  mapGroup = map.mMonsterGroups[i];
        CMonsterGroup2& group = monsterGroups[i];

        group.update(mapGroup.mPos, mapGroup.getType());
    }
}

//---------------------------------------------------------------------------------------------
void CMonster::update(CMonsterGroup2& group, CVec2 mapPos, uint8_t type)
{
    displayAttackTimer.update();

    switch (group.state)
    {
        case eMonsterState_Wandering:
            wander();
            break;

        case eMonsterState_Fight:
            fight(mapPos, type);
            break;
    }
}

//---------------------------------------------------------------------------------------------
void CMonster::wander()
{

}

//---------------------------------------------------------------------------------------------
void CMonster::initNextAttack(uint8_t type)
{
    int minTime = monsters.monstersDatas[type].timeBetweenAttack;
    int nextTime = minTime + (rand() % 40) - 10;

    if (minTime > 150)
        nextTime += (rand() % 80) - 20;

    nextAttackTimer.set(nextTime, true);
}

//---------------------------------------------------------------------------------------------
void CMonster::fight(CVec2 mapPos, uint8_t type)
{
    if (nextAttackTimer.update() == true)
    {
        QString soundName = monsters.monstersDatas[type].attackSound;

        if (soundName.isEmpty() == false)
            sound->play(mapPos, soundName.toLocal8Bit().constData());

        int attackTime = monsters.monstersDatas[type].attackDisplayTime;
        displayAttackTimer.set(attackTime, true);
        initNextAttack(type);
    }
}

//---------------------------------------------------------------------------------------------
CMonsterGroup2::CMonsterGroup2()
{
    state = eMonsterState_Wandering;
}

//---------------------------------------------------------------------------------------------
void CMonsterGroup2::update(CVec2 mapPos, uint8_t type)
{
    // met à jour tous les monstres du groupe
    for (int i = 0; i < 4; ++i)
        if (monsters[i].pos != eMonsterPosNone)
            monsters[i].update(*this, mapPos, type);

    // gère le comportement du groupe pour chaque état
    switch (state)
    {
        case eMonsterState_Wandering:
            wander(mapPos, type);
            break;

        case eMonsterState_Fight:
            fight(mapPos, type);
            break;
    }
}

//---------------------------------------------------------------------------------------------
void CMonsterGroup2::wander(CVec2 mapPos, uint8_t type)
{
    // transition en état de combat ?
    CVec2   dist = player.pos - mapPos;
    bool    isNearPlayer = (dist.x == 0 && ABS(dist.y) == 1) || (dist.y == 0 && ABS(dist.x) == 1);

    if (isNearPlayer == true)
    {
        state = eMonsterState_Fight;

        // direction vers le joueur
        EMonsterDir dir;

        if (dist.x > 0)
            dir = eMonsterDirRight;
        else if (dist.x < 0)
            dir = eMonsterDirLeft;
        else if (dist.y > 0)
            dir = eMonsterDirDown;
        else
            dir = eMonsterDirUp;

        // initialise chaque monstre du groupe
        for (int i = 0; i < 4; ++i)
            if (monsters[i].pos != eMonsterPosNone)
            {
                monsters[i].dir = dir;
                monsters[i].initNextAttack(type);
            }
    }
}

//---------------------------------------------------------------------------------------------
void CMonsterGroup2::fight(CVec2 mapPos, uint8_t type)
{
    // transition en mode ballade ?
    CVec2   dist = player.pos - mapPos;
    bool    isNearPlayer = (dist.x == 0 && ABS(dist.y) == 1) || (dist.y == 0 && ABS(dist.x) == 1);

    if (isNearPlayer == false)
    {
        state = eMonsterState_Wandering;
    }
}
